// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AI32SSA/driver/ioctl.c $
// $Rev: 51927 $
// $Date: 2022-10-20 17:44:12 -0500 (Thu, 20 Oct 2022) $

// 16AI32SSA: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START			D13		// BCTLR
#define	AUTO_CAL_PASS			D14		// BCTLR
#define	INIT_START				D15		// BCTLR

#define	IRQ0_INIT_DONE			0x0000	// ICR
#define	IRQ0_AUTO_CAL_DONE		0x0001	// ICR
#define	IRQ0_MASK				0x0007	// ICR
#define	IRQ0_ACTIVE				0x0008	// ICR
#define	IRQ1_ACTIVE				0x0080	// ICR



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	switch (arg[0])
	{
		default:							arg[0]	= AI32SSA_IOCTL_QUERY_ERROR;		break;
		case AI32SSA_QUERY_AUTO_CAL_MS:		arg[0]	= dev->cache.auto_cal_ms;			break;
		case AI32SSA_QUERY_CHANNEL_MAX:		arg[0]	= dev->cache.channels_max;			break;
		case AI32SSA_QUERY_CHANNEL_QTY:		arg[0]	= dev->cache.channel_qty;			break;
		case AI32SSA_QUERY_COUNT:			arg[0]	= AI32SSA_IOCTL_QUERY_LAST + 1;		break;
		case AI32SSA_QUERY_DEVICE_TYPE:		arg[0]	= dev->board_type;					break;
		case AI32SSA_QUERY_FIFO_SIZE:		arg[0]	= dev->cache.fifo_size;				break;
		case AI32SSA_QUERY_FSAMP_MAX:		arg[0]	= dev->cache.fsamp_max;				break;
		case AI32SSA_QUERY_FSAMP_MIN:		arg[0]	= dev->cache.fsamp_min;				break;
		case AI32SSA_QUERY_INIT_MS:			arg[0]	= dev->cache.initialize_ms;			break;
		case AI32SSA_QUERY_MASTER_CLOCK:	arg[0]	= dev->cache.master_clock;			break;
		case AI32SSA_QUERY_NRATE_MASK:		arg[0]	= dev->cache.rate_gen_nrate_mask;	break;
		case AI32SSA_QUERY_NRATE_MAX:		arg[0]	= dev->cache.rate_gen_nrate_max;	break;
		case AI32SSA_QUERY_NRATE_MIN:		arg[0]	= dev->cache.rate_gen_nrate_min;	break;
		case AI32SSA_QUERY_RATE_GEN_QTY:	arg[0]	= dev->cache.rate_gen_qty;			break;
		case AI32SSA_QUERY_V_RANGE_HIGH:	arg[0]	= dev->cache.v_range_high;			break;
	}

	return(0);
}



//*****************************************************************************
static int _init_start(dev_data_t* dev, void* arg)
{
	// Initiate initialization.
	os_reg_mem_tx_u32(NULL, dev->vaddr.gsc_bctlr_32, INIT_START);
	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	u32			bctlr;
	int			i;
	int			mask;
	long		ms			= dev->cache.initialize_ms + 5000;
	long		ms_total	= ms;
	os_sem_t	sem;
	int			ret			= 0;
	int			tmp;
	int			value;
	gsc_wait_t	wt;

	if (dev->irq.opened)
	{
		ms_total	*= 2;

		// Safely select the Initialize Done interrupt.
		mask	= IRQ0_MASK | IRQ0_ACTIVE | IRQ1_ACTIVE;
		value	= IRQ0_INIT_DONE;
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, value, mask);

		// Wait for the local interrupt.
		os_sem_create(&sem);	// dummy, required for wait operations.
		memset(&wt, 0, sizeof(wt));
		wt.flags		= GSC_WAIT_FLAG_INTERNAL;
		wt.gsc			= AI32SSA_WAIT_GSC_INIT_DONE;
		wt.timeout_ms	= ms;
		ret				= gsc_wait_event(dev, &wt, _init_start, NULL, &sem);
		os_sem_destroy(&sem);

		if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			ret	= ret ? ret : -ETIMEDOUT;
			printf(	"%s: INITILIZE DONE IRQ TIMED OUT AFTER %ld ms.\n",
					dev->model,
					(long) ms);
		}
	}

	// Manually wait for completion as the IRQ wait may have ended early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bctlr_32, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bctlr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (bctlr & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	// Wait for settling.
	os_time_sleep_ms(100);
	return(ret);
}



//*****************************************************************************
static int _auto_cal_start(dev_data_t* dev, void* arg)
{
	// Initiate auto-calibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bctlr_32, AUTO_CAL_START, AUTO_CAL_START);
	return(0);
}



//*****************************************************************************
static int _auto_calibrate(dev_data_t* dev, void* arg)
{
	u32			bctlr;
	u32			mask;
	long		ms		= dev->cache.auto_cal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Auto-Cal Done interrupt.
	mask	= IRQ0_MASK | IRQ0_ACTIVE | IRQ1_ACTIVE;
	value	= IRQ0_AUTO_CAL_DONE;
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AI32SSA_WAIT_GSC_AUTO_CAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _auto_cal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bctlr_32, AUTO_CAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bctlr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (bctlr & AUTO_CAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if ((reg & AUTO_CAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	if (ret == 0)
	{
		// Wait for settling.
		os_time_sleep_ms(100);
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_sts(dev_data_t* dev, s32* arg)
{
	u32	reg;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (reg & AUTO_CAL_START)
		arg[0]	= AI32SSA_AUTO_CAL_STS_ACTIVE;
	else if (reg & AUTO_CAL_PASS)
		arg[0]	= AI32SSA_AUTO_CAL_STS_PASS;
	else
		arg[0]	= AI32SSA_AUTO_CAL_STS_FAIL;

	return(0);
}



//*****************************************************************************
static int _ain_buf_clear(dev_data_t* dev, void* arg)
{
	int	ret;

	// This action clears the buffer and the over and under run flags.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_ibcr_32, D18, D18);

	// The bit should clear within 200ns.
	ret	= gsc_poll_u32(dev, 2, dev->vaddr.gsc_ibcr_32, D18, 0);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_level(dev_data_t* dev, s32* arg)
{
	arg[0]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bufsr_32);
	arg[0]	&= 0x7FFFF;
	return(0);
}



//*****************************************************************************
static int _ain_buf_overflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AIN_BUF_OVERFLOW_NO,
		AI32SSA_AIN_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_thr_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x3FFFF, dev->vaddr.gsc_ibcr_32, 17, 0);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_thr_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AIN_BUF_THR_STS_CLEAR,
		AI32SSA_AIN_BUF_THR_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibcr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_underflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AIN_BUF_UNDERFLOW_NO,
		AI32SSA_AIN_BUF_UNDERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ain_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AIN_MODE_SYSTEM,
		AI32SSA_AIN_MODE_ZERO,
		AI32SSA_AIN_MODE_VREF,
		-1	// terminate list
	};

	int	ret;
	s32	v	= arg[0];

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 2, 0);

	if ((ret == 0) && (v != -1))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _ain_range(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;
	s32	v	= arg[0];

	if (dev->cache.v_range_high)
	{
		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case -1:

				reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

				switch (reg & 0x30)
				{
					case 0x00:	arg[0]	= AI32SSA_AIN_RANGE_2_5V;	break;
					case 0x10:	arg[0]	= AI32SSA_AIN_RANGE_5V;		break;
					default:
					case 0x20:
					case 0x30:	arg[0]	= AI32SSA_AIN_RANGE_10V;	break;
				}

				break;

			case AI32SSA_AIN_RANGE_2_5V:
			case AI32SSA_AIN_RANGE_5V:
			case AI32SSA_AIN_RANGE_10V:

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, (arg[0] - 2) << 4, 0x30);
				break;
		}
	}
	else
	{
		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case -1:

				reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

				switch (reg & 0x30)
				{
					case 0x00:	arg[0]	= AI32SSA_AIN_RANGE_0_625V;	break;
					case 0x10:	arg[0]	= AI32SSA_AIN_RANGE_1_25V;	break;
					default:
					case 0x20:
					case 0x30:	arg[0]	= AI32SSA_AIN_RANGE_2_5V;	break;
				}

				break;

			case AI32SSA_AIN_RANGE_0_625V:
			case AI32SSA_AIN_RANGE_1_25V:
			case AI32SSA_AIN_RANGE_2_5V:

				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, arg[0] << 4, 0x30);
				break;
		}
	}

	if ((ret == 0) && (v == -1))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _aux_0_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AUX_MODE_DISABLE,
		AI32SSA_AUX_MODE_INPUT,
		AI32SSA_AUX_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 1, 0);
	return(ret);
}



//*****************************************************************************
static int _aux_1_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AUX_MODE_DISABLE,
		AI32SSA_AUX_MODE_INPUT,
		AI32SSA_AUX_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 3, 2);
	return(ret);
}



//*****************************************************************************
static int _aux_in_pol(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AUX_IN_POL_LO_2_HI,
		AI32SSA_AUX_IN_POL_HI_2_LO,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _aux_noise(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AUX_NOISE_LOW,
		AI32SSA_AUX_NOISE_HIGH,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _aux_out_pol(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AUX_OUT_POL_HI_PULSE,
		AI32SSA_AUX_OUT_POL_LOW_PULSE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _aux_sync_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_AUX_SYNC_ENABLE_NO,
		AI32SSA_AUX_SYNC_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 7, 7);
	return(ret);
}



//*****************************************************************************
static int _burst_busy(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_BURST_BUSY_IDLE,
		AI32SSA_BURST_BUSY_ACTIVE,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 7, 7);
	return(ret);
}



//*****************************************************************************
static int _burst_size(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFF, dev->vaddr.gsc_bsizr_32, 19, 0);
	return(ret);
}



//*****************************************************************************
static int _burst_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_BURST_SRC_DISABLE,
		AI32SSA_BURST_SRC_RBG,
		AI32SSA_BURST_SRC_EXT,
		AI32SSA_BURST_SRC_BCR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 9, 8);
	return(ret);
}



//*****************************************************************************
static int _chan_active(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_CHAN_ACTIVE_SINGLE,
		AI32SSA_CHAN_ACTIVE_0_1,
		AI32SSA_CHAN_ACTIVE_0_3,
		AI32SSA_CHAN_ACTIVE_0_7,
		AI32SSA_CHAN_ACTIVE_0_15,
		AI32SSA_CHAN_ACTIVE_0_3X,
		AI32SSA_CHAN_ACTIVE_RANGE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 2, 0);
	return(ret);
}



//*****************************************************************************
static int _chan_first(dev_data_t* dev, void* arg)
{
	u32	last;
	u32	reg;
	int	ret;

	reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acar_32);
	last	= GSC_FIELD_DECODE(reg, 15, 8);
	ret		= gsc_s32_range_reg(dev, arg, 0, last, dev->vaddr.gsc_acar_32, 7, 0);
	return(ret);
}



//*****************************************************************************
static int _chan_last(dev_data_t* dev, void* arg)
{
	u32	first;
	u32	last;
	u32	reg;
	int	ret;

	last	= dev->cache.channel_qty - 1;
	reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acar_32);
	first	= GSC_FIELD_DECODE(reg, 7, 0);
	ret		= gsc_s32_range_reg(dev, arg, first, last, dev->vaddr.gsc_acar_32, 15, 8);
	return(ret);
}



//*****************************************************************************
static int _chan_single(dev_data_t* dev, void* arg)
{
	u32	last;
	int	ret;

	last	= dev->cache.channel_qty - 1;
	ret		= gsc_s32_range_reg(dev, arg, 0, last, dev->vaddr.gsc_sscr_32, 17, 12);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_DATA_FORMAT_2S_COMP,
		AI32SSA_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _data_packing(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_DATA_PACKING_DISABLE,
		AI32SSA_DATA_PACKING_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _dmdma_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_DMDMA_MODE_THRESHOLD,
		AI32SSA_DMDMA_MODE_AUTONOMOUS,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _input_sync(dev_data_t* dev, void* arg)
{
	unsigned long	ms_limit;
	int				ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, D12, D12);

	// Wait for the operation to complete.

	if (dev->io.rx.timeout_s)
		ms_limit	= dev->io.rx.timeout_s * 1000;
	else
		ms_limit	= 1000;

	ret	= gsc_poll_u32(dev, ms_limit, dev->vaddr.gsc_bctlr_32, D12, 0);
	return(ret);
}



//*****************************************************************************
static int _irq0_sel(dev_data_t* dev, s32* arg)
{
	int	ret	= 0;
	u32	val;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		val		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
		arg[0]	= GSC_FIELD_DECODE(val, 2, 0);
	}
	else
	{
		// Validate the option value passed in.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case AI32SSA_IRQ0_INIT_DONE:
			case AI32SSA_IRQ0_AUTO_CAL_DONE:
			case AI32SSA_IRQ0_SYNC_START:
			case AI32SSA_IRQ0_SYNC_DONE:
			case AI32SSA_IRQ0_BURST_START:
			case AI32SSA_IRQ0_BURST_DONE:

				break;
		}

		if (ret == 0)
		{
			// Clear the IRQ0 status bit and apply the new setting.
			val	= (dev->cache.fw_ver >= 0x003) ? (arg[0] | 0x88) : (arg[0]);
			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, val, 0x8F);
		}
	}

	return(ret);
}



//*****************************************************************************
static int _irq1_sel(dev_data_t* dev, s32* arg)
{
	int	ret	= 0;
	u32	val;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		val		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
		arg[0]	= GSC_FIELD_DECODE(val, 6, 4);
	}
	else
	{
		// Validate the option value passed in.

		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case AI32SSA_IRQ1_NONE:
			case AI32SSA_IRQ1_IN_BUF_THR_L2H:
			case AI32SSA_IRQ1_IN_BUF_THR_H2L:
			case AI32SSA_IRQ1_IN_BUF_OVR_UNDR:

				break;
		}

		if (ret == 0)
		{
			// Clear the IRQ1 status bit and apply the new setting.
			val	= (dev->cache.fw_ver >= 0x003) ? ((arg[0] << 4) | 0x88) : (arg[0] << 4);
			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, val, 0xF8);
		}
	}

	return(ret);
}



//*****************************************************************************
static int _rag_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_GEN_ENABLE_NO,
		AI32SSA_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ragr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _rag_nrate(dev_data_t* dev, s32* arg)
{
	int	ret;
	s32	v	= arg[0];

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.rate_gen_nrate_min,
			dev->cache.rate_gen_nrate_max,
			dev->vaddr.gsc_ragr_32,
			15,
			0);

	if ((ret == 0) && (v != -1))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _rbg_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_RBG_CLK_SRC_MASTER,
		AI32SSA_RBG_CLK_SRC_RAG,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _rbg_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_GEN_ENABLE_NO,
		AI32SSA_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_rbgr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _rbg_nrate(dev_data_t* dev, s32* arg)
{
	int	ret;
	s32	v	= arg[0];

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.rate_gen_nrate_min,
			dev->cache.rate_gen_nrate_max,
			dev->vaddr.gsc_rbgr_32,
			15,
			0);

	if ((ret == 0) && (v != -1))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



//*****************************************************************************
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		GSC_IO_MODE_DMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.rx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _rx_io_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_IO_OVERFLOW_IGNORE,
		AI32SSA_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AI32SSA_IO_TIMEOUT_MIN,
			AI32SSA_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _rx_io_underflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_IO_UNDERFLOW_IGNORE,
		AI32SSA_IO_UNDERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.underflow_check);
	return(ret);
}



//*****************************************************************************
static int _samp_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_SAMP_CLK_SRC_RAG,
		AI32SSA_SAMP_CLK_SRC_RBG,
		AI32SSA_SAMP_CLK_SRC_EXT,
		AI32SSA_SAMP_CLK_SRC_BCR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 4, 3);
	return(ret);
}



//*****************************************************************************
static int _scan_marker(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSA_SCAN_MARKER_DISABLE,
		AI32SSA_SCAN_MARKER_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _scan_marker_val(dev_data_t* dev, u32* arg)
{
	u32	lwr;
	u32	upr;

	if ((s32) arg[0] == (s32) -1)
	{
		upr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_smuwr_32);
		lwr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_smlwr_32);
		arg[0]	= ((upr & 0xFFFF) << 16) | (lwr & 0xFFFF);
	}
	else
	{
		lwr	= arg[0] & 0xFFFF;
		upr	= arg[0] >> 16;
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_smuwr_32, upr, 0xFFFF);
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_smlwr_32, lwr, 0xFFFF);
	}

	return(0);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AI32SSA_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ AI32SSA_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ AI32SSA_IOCTL_REG_MOD,			(void*) gsc_reg_mod_ioctl		},
	{ AI32SSA_IOCTL_QUERY,				(void*) _query					},
	{ AI32SSA_IOCTL_INITIALIZE,			(void*) initialize_ioctl		},
	{ AI32SSA_IOCTL_AUTO_CALIBRATE,		(void*) _auto_calibrate			},
	{ AI32SSA_IOCTL_AUTO_CAL_STS,		(void*) _auto_cal_sts			},
	{ AI32SSA_IOCTL_AIN_BUF_CLEAR,		(void*) _ain_buf_clear			},
	{ AI32SSA_IOCTL_AIN_BUF_LEVEL,		(void*) _ain_buf_level			},
	{ AI32SSA_IOCTL_AIN_BUF_OVERFLOW,	(void*) _ain_buf_overflow		},
	{ AI32SSA_IOCTL_AIN_BUF_THR_LVL,	(void*) _ain_buf_thr_lvl		},
	{ AI32SSA_IOCTL_AIN_BUF_THR_STS,	(void*) _ain_buf_thr_sts		},
	{ AI32SSA_IOCTL_AIN_BUF_UNDERFLOW,	(void*) _ain_buf_underflow		},
	{ AI32SSA_IOCTL_AIN_MODE,			(void*) _ain_mode				},
	{ AI32SSA_IOCTL_AIN_RANGE,			(void*) _ain_range				},
	{ AI32SSA_IOCTL_AUX_0_MODE,			(void*) _aux_0_mode				},
	{ AI32SSA_IOCTL_AUX_1_MODE,			(void*) _aux_1_mode				},
	{ AI32SSA_IOCTL_AUX_IN_POL,			(void*) _aux_in_pol				},
	{ AI32SSA_IOCTL_AUX_NOISE,			(void*) _aux_noise				},
	{ AI32SSA_IOCTL_AUX_OUT_POL,		(void*) _aux_out_pol			},
	{ AI32SSA_IOCTL_AUX_SYNC_ENABLE,	(void*) _aux_sync_enable		},
	{ AI32SSA_IOCTL_BURST_BUSY,			(void*) _burst_busy				},
	{ AI32SSA_IOCTL_BURST_SIZE,			(void*) _burst_size				},
	{ AI32SSA_IOCTL_BURST_SRC,			(void*) _burst_src				},
	{ AI32SSA_IOCTL_CHAN_ACTIVE,		(void*) _chan_active			},
	{ AI32SSA_IOCTL_CHAN_FIRST,			(void*) _chan_first				},
	{ AI32SSA_IOCTL_CHAN_LAST,			(void*) _chan_last				},
	{ AI32SSA_IOCTL_CHAN_SINGLE,		(void*) _chan_single			},
	{ AI32SSA_IOCTL_DATA_FORMAT,		(void*) _data_format			},
	{ AI32SSA_IOCTL_DATA_PACKING,		(void*) _data_packing			},
	{ AI32SSA_IOCTL_DMDMA_MODE,			(void*) _dmdma_mode				},
	{ AI32SSA_IOCTL_INPUT_SYNC,			(void*) _input_sync				},
	{ AI32SSA_IOCTL_IRQ0_SEL,			(void*) _irq0_sel				},
	{ AI32SSA_IOCTL_IRQ1_SEL,			(void*) _irq1_sel				},
	{ AI32SSA_IOCTL_RAG_ENABLE,			(void*) _rag_enable				},
	{ AI32SSA_IOCTL_RAG_NRATE,			(void*) _rag_nrate				},
	{ AI32SSA_IOCTL_RBG_CLK_SRC,		(void*) _rbg_clk_src			},
	{ AI32SSA_IOCTL_RBG_ENABLE,			(void*) _rbg_enable				},
	{ AI32SSA_IOCTL_RBG_NRATE,			(void*) _rbg_nrate				},
	{ AI32SSA_IOCTL_RX_IO_ABORT,		(void*) _rx_io_abort			},
	{ AI32SSA_IOCTL_RX_IO_MODE,			(void*) _rx_io_mode				},
	{ AI32SSA_IOCTL_RX_IO_OVERFLOW,		(void*) _rx_io_overflow			},
	{ AI32SSA_IOCTL_RX_IO_TIMEOUT,		(void*) _rx_io_timeout			},
	{ AI32SSA_IOCTL_RX_IO_UNDERFLOW,	(void*) _rx_io_underflow		},
	{ AI32SSA_IOCTL_SAMP_CLK_SRC,		(void*) _samp_clk_src			},
	{ AI32SSA_IOCTL_SCAN_MARKER,		(void*) _scan_marker			},
	{ AI32SSA_IOCTL_SCAN_MARKER_VAL,	(void*) _scan_marker_val		},
	{ AI32SSA_IOCTL_WAIT_CANCEL,		(void*) gsc_wait_cancel_ioctl	},
	{ AI32SSA_IOCTL_WAIT_EVENT,			(void*) gsc_wait_event_ioctl	},
	{ AI32SSA_IOCTL_WAIT_STATUS,		(void*) gsc_wait_status_ioctl	},
	{ -1, NULL }
};


